import sys

# Test the output and error redirection
sys.stdout.write("Hello World!\n")
sys.stderr.write("Hello Error!\n")

# Test creating an output file
with open("test.out", 'w') as f:
    f.write("This is only a test.\n")






class Namespace(object):
    def __init__(self, name=None, **kwargs):
        self.__dict__.update(kwargs)
        
        if name != None:
            register(name, self)

NamespaceRegistry = {}

def register(name, namespace):
    NamespaceRegistry[name] = namespace

def nsupdate(name, attr, value):
    setattr(NamespaceRegistry[name], attr, value)

__all__ = [
    "Namespace",
    "NamespaceRegistry",
    "nsupdate",
]



"""
    experipy.exp
    ~~~~~~~~~~~~

    This module provides the Experiment class for running compositions in the
    grammar, as well as the Exp Namespace for controlling and configuring 
    Experiment behavior.
"""
import shutil
import sys

from datetime   import datetime
from os         import chmod, makedirs, path
from subprocess import call
from time       import time

from .grammar   import Element
from .system    import Cd, Cp, Mkdir, Rm
from .utils     import Namespace

Exp = Namespace("Exp",
    runsh   = "run.sh",
    shebang = "#!/bin/bash",
    rundir  = "/tmp",
    defname = "exp",
    out     = "raw.out",
    err     = "raw.err",
    timing  = "harness_time.out",
)


class ExpError(Exception):
    pass


class Experiment(object):
    """Experiment objects perform the generation and execution of runscripts.
    
    Once a composition has been specified in the grammar, wrapping it in an
    Experiment allows the user to generate a shell scripts as a string using
    the make_runscript method. The run and queue methods provide mechanisms for
    executing the generated scripts.

    Parameters
    ----------
    cmd : experipy.Element
        A composition of experipy Elements such as Executable and Group, which
        defines the behavior the user wishes the Experiment to perform.
    expname : str
        A name to be used for identifying the experiment. Defaults to 
        Exp.defname, which defaults to "exp"
    destdir : str
        An optional path to a directory where the results from running the
        experiment should be stored. If None, expname will be used.

    """

    def __init__(self, cmd, expname=Exp.defname, destdir=None):
        if not isinstance(cmd, Element):
            raise ExpError("'{}' is not an instance of Element".format(cmd))

        self.cmd     = cmd
        self.expname = expname

        if not destdir:
            self.destdir = path.abspath(expname)
        else:
            self.destdir = path.abspath(destdir)
    

    def make_runscript(self, preamble=Exp.shebang, rm_rundir=True):
        """Create a string containing a complete shell script.

        Parameters
        ----------
        preamble : str
            The first line(s) of the runscript. Defaults to Exp.shebang, which
            defaults to "#!/bin/bash".
        rm_rundir : bool
            If True, a line deleting the experiment's working directory will
            be added to the end of the script. Defaults to True.

        Returns
        -------
        str
            A run script as described by the composition provided to the 
            Experiment.
        """

        # Name of the temporary directory for the experiment
        rundir = path.join(Exp.rundir, self.expname + "." + str(int(time())))
        
        # Start with the preamble
        scriptstr = preamble + "\n\n"
        
        # Collect experiment input files
        scriptstr += "# Experiment setup\n"

        scriptstr += str(Mkdir(rundir, make_parents=True)) + "\n"
        scriptstr += str(Cd(rundir)) + "\n"

        for infile in self.cmd.inputs:
            scriptstr += str(Cp(path.abspath(infile), ".")) + "\n"
        
        # Execute the experiment components
        scriptstr += "\n# Run experiment\n"
        scriptstr += str(self.cmd) + "\n\n"
        
        # Exfill the experiment output files and clean up the rundir if needed
        scriptstr += "# Collect output files and clean up\n"
        for outfile in self.cmd.outputs:
            scriptstr += str(Cp(outfile, self.destdir)) + "\n"
        
        if rm_rundir:
            scriptstr += str(Rm(rundir)) + "\n"

        return scriptstr

    
    def run(self, rm_rundir=True):
        """Execute the experiment as a subprocess of the current process.

        Generates a run script, writes that script to the results directory,
        and then executes the script as a subprocess of the current process.
        The time the script takes to execute, including loader time, is 
        recorded. This function blocks until the experiment is complete.

        Parameters
        ----------
        rm_rundir : bool
            If True, the directory created for running the experiment will be
            deleted at the end of the experiment. Defaults to True.
        """

        # Create the results directory, deleting any previous contents
        if path.exists(self.destdir):
            shutil.rmtree(self.destdir)
        makedirs(self.destdir)
        
        # Open the output, error and timing file handles for call
        out = open(path.join(self.destdir, Exp.out), 'w')
        err = open(path.join(self.destdir, Exp.err), 'w')
        timing = open(path.join(self.destdir, Exp.timing), 'w')

        # Write the runscript
        fname  = path.join(self.destdir, Exp.runsh)
        with open(fname, "w") as f:
            f.write(self.make_runscript(rm_rundir=rm_rundir))
        chmod(fname, 0755)

        # Execute call and time the result
        start = datetime.now()
        call(fname, stdout=out, stderr=err)
        runtime = datetime.now() - start
        timing.write(str(runtime)+"\n")

        # Clean up and close file descriptors
        out.close()
        err.close()
        timing.close()
        

    def queue(self, nodes=1, ppn=1, mem="4096m", wtime="480:00:00"):
        """Submit the experiment to a job queuing system as a PBS script.
        
        Generates a script with a PBS script header, writes the script to the 
        results directory, and then submits it to the job queuing system by
        running the command qsub as a subprocess.

        Parameters
        ----------
        nodes : int
            The number of nodes to request in the PBS script. Defaults to 1.
        ppn : int
            The number of processors per node to request in the PBS script.
            Defaults to 1.
        mem : str
            The amount of memory to request in the PBS script. Defaults to 
            "4096m".
        wtime : str
            The amount of wall time to request in the PBS script.
        """
       
        # Write the PBS script preamble
        pbsheader = (Exp.shebang
            + "\n#PBS -N {name}\n"
            + "#PBS -l nodes={nodes}:ppn={ppn},mem={mem}\n"
            + "#PBS -l walltime={wtime}\n"
            + "#PBS -o {qout}\n#PBS -e {qerr}"
            + ""
        ).format(name=self.expname, nodes=nodes, ppn=ppn, mem=mem, wtime=wtime,
            qout=path.join(self.destdir, Exp.out), 
            qerr=path.join(self.destdir, Exp.err)
        )
        
        # Create the results directory, deleting any previous contents
        if path.exists(self.destdir):
            shutil.rmtree(self.destdir)
        makedirs(self.destdir)
        
        # Write the runscript
        fname  = path.join(self.destdir, Exp.runsh)
        with open(fname, "w") as f:
            f.write(self.make_runscript(
                preamble=pbsheader, 
                rm_rundir=rm_rundir
            ))
        
        chmod(fname, 0755)
        
        # Submit to the queue
        call(["qsub", fname])



from os                 import path

from experipy.utils     import Namespace
from experipy.grammar   import Executable

from .environment       import Env

########## Constants and helper functions ##########

Spec = Namespace("Spec",
    path = "benchmarks/cpu2006",
)

def specpath():
    return path.join(Env.projdir, Spec.path)

####################################################

class SpecBench(Executable):
    pass



from __future__ import absolute_import

from . import (
    exp,
    grammar,
    system,
    utils,
)


__version__     = "0.1.3"

__title__       = "experipy"
__description__ = "A framework for writing and running Computational Science experiments"
__uri__         = "https://github.com/Elemnir/experipy"

__author__      = "Adam Howard"
__email__       = "ahoward0920@gmail.com"

__license__     = "BSD 3-clause"
__copyright__   = "Copyright (c) 2016 Adam Howard"


Experiment  = exp.Experiment
Namespace   = utils.Namespace

__all__ = [
    "exp",
    "grammar",
    "system",
    "utils",
    "Experiment",
    "Namespace",
]



from os                 import path

from experipy.utils     import Namespace
from experipy.grammar   import (Executable, Wrapper, Group, 
                                GrammarViolation, tokens)

from .environment       import Env

########## Constants and helper functions ##########

Pin = Namespace("Pin",
    path = "pin-2.14",
    exe  = "pin.sh",
)

def pinpath():
    return path.join(Env.projdir, Pin.path)

####################################################

class PinTool(Wrapper):
    def __init__(self, tool, target, popts=[], topts=[], **kwargs):
        super(PinTool, self).__init__(
            Executable(
                path.join(pinpath(), Pin.exe),
                popts + ["-t "+tool] + topts + ["--", tokens.wrapped]
            ), target, **kwargs
        )

class Memtracer(PinTool):
    def __init__(self, target, popts=["-follow-execv"], topts=[], **kwargs):
        super(Memtracer, self).__init__(
            path.join(pinpath(), "source/tools/memtracer/obj-intel64/", "memtracer.so"),
            target, popts, topts, **kwargs
        )




from os import path

from experipy.system import PythonScript
from experipy        import Experiment

if __name__ == "__main__":
    testscript = path.join(path.dirname(__file__), "test.py")
    Experiment(PythonScript(testscript, outputs=["test.out"]), 
        "test",
        path.join(path.dirname(__file__), "results/")
    ).run(rm_rundir=False)



"""
    experipy.grammar
    ~~~~~~~~~~~~~~~~

    This module provides the core elements which compose the Experipy grammar:
    Executables, Wrappers, Pipelines, and Groups. These elements facilitate 
    specifying programs to execute as well as the files they depend on. 
"""
import re

from .utils import Namespace

tokens = Namespace(
    wrapped = r"[[wrapped]]",
)

class GrammarViolation(Exception):
    pass


class Element(object):
    """The Element class forms the grammar's base class.

    Parameters
    ----------
    inputs : list
        A list of strings which are the names of files that the Element relies
        on for input. These will be copied to the run directory when an 
        Experiment is used to run the Element.
    outputs: list
        A list of strings which are the names of files that the Element is 
        expected to generate as output. These will be copied from the run
        directory when an Experiment is used to run the Element.
    """

    def __init__(self, **kwargs):
        self.inputs = kwargs.get("inputs", [])
        self.outputs = kwargs.get("outputs", [])


class Executable(Element):
    """Executable objects should represent a single program and its arguments.

    Parameters
    ----------
    prog : str
        The name of the program executable.
    opts : list
        A list of command line options to pass to the program. Defaults to
        an empty list if not provided.
    wait : bool
        If False, a '&' will be appended to the argument list, indicating to
        the shell that it should background the program instead of blocking
        on it. Defaults to True.
    """

    def __init__(self, prog, opts=None, wait=True, **kwargs):
        super(Executable, self).__init__(**kwargs)
        self.prog = prog
        self.opts = []
        if opts != None:
            self.opts.extend(opts)

        if wait == False:
            self.opts.append("&")

    def __str__(self):
        """Render the Executable as it will appear in the shell script."""
        return "{0} {1}".format(self.prog, " ".join(map(str,self.opts)))


class Wrapper(Element):
    """Wrapper objects allow one program to wrap another.

    Provides a way to specify programs such as GDB or Valgrind, which wrap
    around another program to alter or observe its execution.

    Parameters
    ----------
    exe : experipy.Executable
        The wrapping Executable. Must provide a '[[wrapped]]' option which 
        indicates where the wrapped Element should be placed when rendering
        the script.
    wrapped : experipy.Executable or experipy.Wrapper
        The wrapped Executable or Wrapper. Inputs and outputs specified to 
        exe and wrapped will also be included in the resultant object's inputs 
        and outputs.
    wait : bool
        If False, a '&' will be appended to the argument list, indicating to
        the shell that it should background the program instead of blocking on
        it. Defaults to True.

    """

    def __init__(self, exe, wrapped, wait=True, **kwargs):
        if not isinstance(exe, Executable):
            raise GrammarViolation("exe must be an instance of Executable")
        elif not (isinstance(wrapped, Executable) or isinstance(wrapped, Wrapper)):
            raise GrammarViolation("wrapped must be an instance of Executable or Wrapper")
        elif tokens.wrapped not in exe.opts:
            raise GrammarViolation("exe must have a '{}' option".format(tokens.wrapped))
        
        inputs = kwargs.get("inputs", [])
        if wrapped.inputs:
            inputs.extend(wrapped.inputs)
        kwargs["inputs"] = inputs

        outputs = kwargs.get("outputs", [])
        if wrapped.outputs:
            outputs.extend(wrapped.outputs)
        kwargs["outputs"] = outputs
        
        if wait == False:
            exe.opts.append("&")

        super(Wrapper, self).__init__(**kwargs)
        self.exe = exe
        self.wrapped = wrapped

    def __str__(self):
        """Render the Wrapper as it will appear in the shell script."""
        return str(self.exe).replace(tokens.wrapped, str(self.wrapped))


class Pipeline(Element):
    """Pipeline objects allow specification of pipelined workflows.

    A Pipeline takes one or more Element parts, and joins them with a '|'
    operator, indicating to the shell that each part should recieve its input
    from the previous part, and provide its output to the next.

    Parameters
    ----------
    *parts : 
        One or more Executables or Wrappers to be chained together into a 
        pipeline. Inputs and outputs to the individual parts will be included
        in the Pipeline's inputs and outputs.
    """

    def __init__(self, *parts, **kwargs):
        for part in parts:
            if not (isinstance(part, Executable) or isinstance(part, Wrapper)):
                raise GrammarViolation("'{}' is not an instance of Executable or Wrapper".format(part))
        
        inputs = kwargs.get("inputs", [])
        for part in parts:
            if part.inputs:
                inputs.extend(part.inputs)
        kwargs["inputs"] = inputs

        outputs = kwargs.get("outputs", [])
        for part in parts:
            if part.outputs:
                outputs.extend(part.outputs)
        kwargs["outputs"] = outputs
        
        super(Pipeline, self).__init__(**kwargs)
        self.parts = parts

    def __str__(self):
        """Render the Pipeline as it will appear in the shell script."""
        return " | ".join(map(str,self.parts))


class Group(Element):
    """Group objects allow specification of Executables to be run in order.

    In the resultant script, a Group's parts will be included one after another,
    in the order they were specified. Groups should be used when specifying 
    complex experiments involving multiple steps like set up or post-processing,
    or combined with the wait parameter to Executable to specify programs which
    should be run concurrently.

    Parameters
    ----------
    *parts :
        One or more Elements to be placed into the script. Inputs and outputs
        to the individual parts will be included in the Group's inputs and 
        outputs.
    """
    def __init__(self, *parts, **kwargs):
        for part in parts:
            if not (isinstance(part, Element)):
                raise GrammarViolation("'{}' is not an instance of Element".format(part))
        
        inputs = kwargs.get("inputs", [])
        for part in parts:
            if part.inputs:
                inputs.extend(part.inputs)
        kwargs["inputs"] = inputs

        outputs = kwargs.get("outputs", [])
        for part in parts:
            if part.outputs:
                outputs.extend(part.outputs)
        kwargs["outputs"] = outputs
 
        super(Group, self).__init__(**kwargs)
        self.parts = parts

    def __str__(self):
        """Render the Group as it will appear in the shell script."""
        return "\n".join(map(str,self.parts))



"""
    experipy.system
    ~~~~~~~~~~~~~~~

    This module provides a number of system and shell tools for helping to
    specify common tasks within the experipy grammar.
"""
from os         import path

from .grammar   import Executable

class Cd(Executable):
    def __init__(self, dirname):
        super(Cd, self).__init__("cd", [dirname])

class Cp(Executable):
    def __init__(self, target, dest, opts=[]):
        super(Cp, self).__init__("cp", opts + ['-t', dest, target])

class Mkdir(Executable):
    def __init__(self, dirname, make_parents=False):
        opts = []
        if make_parents:
            opts.append("--parents")
        super(Mkdir, self).__init__("mkdir", opts + [dirname])

class Mkfifo(Executable):
    def __init__(self, pipename):
        super(Mkfifo, self).__init__("mkfifo", [pipename])

class Rm(Executable):
    def __init__(self, *files):
        super(Rm, self).__init__("rm", ["-rf"] + list(files))

class Wait(Executable):
    def __init__(self):
        super(Wait, self).__init__("wait")

class PythonScript(Executable):
    def __init__(self, script, sopts=[], pythonexe="python", **kwargs):
        if 'inputs' in kwargs:
            kwargs['inputs'].append(path.abspath(script))
        else:
            kwargs['inputs'] = [path.abspath(script)]

        super(PythonScript, self).__init__(pythonexe, [script] + sopts, **kwargs)

class JavaApp(Executable):
    def __init__(self, jarfile, popts=[], javaexe="java", jopts=[], **kwargs):
        jarfile = path.abspath(jarfile)
        super(JavaApp, self).__init__(javaexe, jopts + ["-jar", jarfile] + popts, **kwargs)



import os
import re

from setuptools import find_packages, setup

NAME = 'experipy'
PACKAGES = find_packages(where='src')
META_PATH = os.path.join("src", "experipy", "__init__.py")
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Natural Language :: English",
    "License :: OSI Approved :: BSD License",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Scientific/Engineering",
    ]
INSTALL_REQUIRES = []

HERE = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with open(os.path.join(HERE, *parts), "r") as f:
        return f.read()

META_FILE = read(META_PATH)

def find_meta(tag):
    meta_match = re.search(
        r"^__{meta}__\s*=\s*['\"]([^'\"]*)['\"]".format(meta=tag),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=tag))


setup(
    name=NAME,
    description=find_meta("description"),
    license=find_meta("license"),
    url=find_meta("uri"),
    version=find_meta("version"),
    author=find_meta("author"),
    author_email=find_meta("email"),
    maintainer=find_meta("author"),
    maintainer_email=find_meta("email"),
    long_description=read("README.rst"),
    packages=PACKAGES,
    package_dir={'':'src'},
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIRES,
)



from os import getenv
from experipy.utils import Namespace

projectdir = getenv("XPS_PROJDIR", "")
if not projectdir:
    projectdir = raw_input("XPS_PROJDIR not set, provide the project directory path\n>>> ")

resultdir = getenv("XPS_VOLDIR", "")
if not resultdir:
    resultdir = raw_input("XPS_VOLDIR not set, provide the result directory path\n>>> ")

Env = Namespace(
    projdir = projectdir,
    resdir  = resultdir,
)



from os                 import path

from experipy.utils     import Namespace
from experipy.grammar   import (Executable, Wrapper, Group, 
                                GrammarViolation, tokens)

from .environment       import Env

########## Constants and helper functions ##########

Marena = Namespace("Marena",
    path = "marena",
    exe  = "marena",
    lib  = "libmarena.so",
)

def marenapath():
    return path.join(Env.projdir, Marena.path)

####################################################

class MarenaRun(Wrapper):
    def __init__(self, wrapped, hotapfile=None, 
                 libmarena=path.join(marenapath(), Marena.lib), 
                 opts=None, **kwargs):
        
        if opts == None:
            opts = []

        if hotapfile:
            opts.extend(["-f", hotapfile])

        super(MarenaRun, self).__init__(
            Executable(
                path.join(marenapath(), Marena.exe),
                ["-l ", libmarena] + opts + ["--", tokens.wrapped]
            ), wrapped, **kwargs
        )

